Post Processing Module to Link Grammar with OPTIC Functions
Donya Quick and Paul Hudak
Last modified: 14-June-2013
For paper: Grammar-Based Automated Music Composition in Haskell

Post processing module to turn Terms into music using Euterpea.

> module PostProc where
> import Euterpea hiding (transpose)
> import GrammarMonad
> import ChordSpaces
> import Data.List
> import System.Random

Intermediate types:

> type Key = (AbsPitch, Mode)
> type RChord = (Key, Dur, CType)
> type TChord = (Key, Dur, AbsChord)
> type TNote = (Key, Dur, AbsPitch)
> type Voice = [TNote]


The goal using these intermediate types is the following:

INPUT	     STEP            OUTPUT         FUNCTION
Seeds -----(grammar)-------> Sentence       gen
Sentence --(mode info)-----> [TChord]       toAbsChords
[TChord] ------------------> [Voice] 		toVoices
[Voice] -------------------> Music Pitch    vsToMusic or vsToMusicI


> toChords :: Term CType -> Mode -> [RChord]
> toChords (NT (Chord d c)) m = [((0, m), d, c)]
> toChords (S ts) m = concatMap (\t -> toChords t m) ts
> toChords (Mod mt x) m = 
>     let (amt, mt') = getMods mt m
>         f = map (\((k,m),d,c) -> ((k+amt,mt'),d,c)) 
>     in  f (toChords x mt') 
> toChords x m = 
>     error ("(toChords) Unable to handle expression: "++showTerm x)

Function to convert into intermediate type [TChord]:

> toAbsChord :: RChord -> TChord
> toAbsChord ((k,m),d,c) = ((k,m), d, t k $ toAs c m)

> toAbsChordNoDim :: RChord -> TChord
> toAbsChordNoDim ((k,m),d,c) = ((k,m), d, t k $ toAsNoDim c m)

> toAbsChords :: Term CType -> Mode -> [TChord]
> toAbsChords ts m = map toAbsChord $ toChords ts m

> toAbsChordsNoDim :: Term CType -> Mode -> [TChord]
> toAbsChordsNoDim ts m = map toAbsChordNoDim$ toChords ts m

Conversion of a single chord to a mode rooted at zero:

> toAs :: CType -> Mode -> [AbsPitch]
> toAs ct m = 
>     let s = getScale m ++ map (+12) s -- fininite scale
>         i = head $ findIndices (==ct) [I, II, III, IV, V, VI, VII] -- can be updated w/enum
>     in  map (s !!) $ map (+i) [0,2,4]

> toAsNoDim :: CType -> Mode -> [AbsPitch]
> toAsNoDim ct m = 
>     let s = getScale m ++ map (+12) s -- fininite scale
>         i = head $ findIndices (==ct) [I, II, III, IV, V, VI, VII] -- can be updated w/enum
>     in  map (s !!) $ fixDim $ map (+i) [0,2,4] where
>         fixDim x = if optEq x [0,3,6] then t (head x) [0,3,7] else x

Fundtions to get information on modes:

> getScale :: Mode -> [AbsPitch]
> getScale Major = [0,2,4,5,7,9,11]
> getScale Minor = [0,2,3,5,7,8,10]

> getMods mt m = 
>     let mts = [M2, M3, M4, M5, M6, M7]
>         i = (head $ findIndices (==mt) mts)+1
>     in  (getScale m !! i, relPat m !! i) where
>    relPat :: Mode -> [Mode]
>    relPat Major = [Major, Minor, Minor, Major, Major, Minor, Minor]
>    relPat Minor = [Minor, Minor, Major, Minor, Minor, Major, Major]

Transposition using a key (to avoid C-major assignment only):

> atTrans :: AbsPitch -> [(Key, Dur, AbsChord)] -> [(Key, Dur, AbsChord)]
> atTrans a = map (\((k,m),d,c) -> ((fixK k a m,m),d, t (a `mod` 12) c)) 

The toCords functon does a similar thing, but returns a CType and 
its key/mode context without performing the conversion to AbsChord.

> ctTrans :: AbsPitch -> [(Key, Dur, CType)] -> [(Key, Dur, CType)]
> ctTrans a = map (\((k,m),d,c) -> ((fixK k a m,m),d,c)) 

> fixK k a Major = (k + a) `mod` 12
> fixK k a Minor = ((k + a) `mod` 12) + 12


Conversion of intermediate type to Music Pitch:

> tChordsToMusic :: [TChord] -> Music Pitch
> tChordsToMusic = line . map f  where
>     f ((k,m),d, as) = chord $ map (\a -> note d (pitch a)) as


============ SPLITTING VOICES APART ===========

The code here places TChords into a form more suitable
for additional musical processing. A Voice is a list of
pitches with duration and key/mode context. 

> toVoices :: [TChord] -> [Voice]
> toVoices ts = 
>     let (ks,ds,ps) = unzip3 ts
>     in  map (\v -> zip3 ks ds v) $ transpose ps

> toNotes :: Voice -> Music Pitch
> toNotes = line . map (\(k,d,p) -> note d (pitch p))

> vsToMusic :: [Voice] -> Music Pitch
> vsToMusic = 
>     chord . map toNotes 

> vsToMusicI :: [InstrumentName] -> [Voice] -> Music Pitch
> vsToMusicI is = 
>     chord . zipWith (\i m -> instrument i m) is . map toNotes


=========

DISPLAY FUNCTIONS

> cShow = latexShow -- preferred format for showing results

> latexShow t ct = show ct ++ "^{" ++ showDur t ++ "}"
> printShow t ct = '(' : show ct ++ " " ++ showDur t ++ ")"

> showTerm :: Term CType -> String
> showTerm t = case t of 
>     Let x a exp -> "Let "++x++"="++show a++" in "++showTerm exp
>     Mod m t -> show m ++ "( "++ showTerm t ++" )"
>     S s -> (concat $ intersperse " " $ map showTerm s)
>     Var v -> v
>     NT (Chord t ct) -> cShow t ct

> showDur :: Dur -> String
> showDur d = if d == sn then "s" else
>             if d == en then "e" else
>             if d == qn then "q" else
>             if d == hn then "h" else
>             if d == wn then "w" else show d


===================

CONSTRAINTS

Let-in expressions impose constraints on the interpretation of
the results as music. These can be viewed as progression-level
constraints, or predicates.

mkCons fundtion looks through the let-in structure of a Term
and produces a progression-level predicate for use with the
ChordSpaces module.

> mkCons :: (Eq a) => Term CType -> Predicate [a]
> mkCons t xs = toCons (findInds [] t) xs where
>     toCons :: (Eq a) => [[(Int, Int)]] -> [a] -> Bool
>     toCons [] xs = True
>     toCons (c:cs) xs = 
>         let f (i,j) = take (j+1-i) $ drop i xs 
>         in  (and $ map (f (head c) == ) $ 
>             map f $ tail c) && toCons cs xs


The findInds function looks through a Term for let-in expressions.
When it finds Let x a exp, it calls findIndsSub on x and exp to 
determine which indices in the sequence are occupied by instances
of x. 

> findInds :: [(String, Term CType)] -> Term CType -> [[(Int, Int)]]
> findInds e t = case t of
>     Mod m t -> findInds e t -- no extra offsets to worry about
>     S [] -> []
>     S (s:ss) -> 
>            let vs1 = findInds e s
>                len = length $ toAbsChords (expand e s) Major
>                vs2 = findInds e (S ss)
>            in  vs1 ++ map (add' len) vs2 where
>                add' y = map (\(a,b) -> (a+y, b+y))
>     Let x a exp -> 
>         let e' = (x, expand e a):e
>             xCases = findIndsSub x e' exp
>         in  xCases : findInds e' a ++ findInds e' exp
>     _ -> [] 

The findIndsSub function looks for instances of a variable and
determines what indices they occupy.

> findIndsSub :: String -> [(String, Term CType)] -> Term CType -> [(Int, Int)]
> findIndsSub x e t = case t of
>     Let x' a exp -> 
>         let a' = expand e a -- to figure out sequence length
>         in  if x'==x then [] -- x has been redefined, no instances to see here
>             else findIndsSub x e a ++ findIndsSub x ((x',a'):e) exp
>     Var x' -> if x/=x' then [] else 
>         let v = maybe (error (x ++ " is undefined")) id $ lookup x e
>             len = length $ toAbsChords v Major
>         in  [(0, len-1)]
>     Mod m t' -> findIndsSub x e t' -- mods don't change dependencies of offsets
>     S [] -> [] -- no instances to worry about
>     S (s:ss) -> 
>         let len = length $ toAbsChords (expand e s) Major
>         in  findIndsSub x e s ++ add' (len) (findIndsSub x e (S ss)) where
>         add' y = map (\(a,b) -> (a+y, b+y))
>     NT v -> [] -- not a variable, so no instances to worry about


> flatten xs = let xs' = flattenRec xs in 
>     if xs' == xs then xs else flatten xs' where
>     flattenRec t = case t of
>         Let x a exp -> Let x (flattenRec a) (flattenRec exp)
>         Var x -> Var x
>         Mod m t -> Mod m (flattenRec t)
>         S [] -> S []
>         S xs -> S $ concatMap stripS $ map flattenRec xs 
>         NT v -> NT v where
>     stripS :: Term a -> [Term a]
>     stripS (S xs) = concatMap stripS xs
>     stripS xs = [xs]




