Search algoritm for use with chord spaces
Donya Quick and Paul Hudak
Last modified: 22-July-2013

Implementation of a search algorithm for traversing chord spaces using
let-in constraints as well as progression level predicates.


> module Search where
> import Data.List
> import ChordSpaces
> import Control.DeepSeq
> import Control.Parallel.Strategies

> type Constraints = [[(Int, Int)]]
> type Index = Int
> type Bound = Int

> printIt = True -- flag for whether to show evidence of search progress
> printIter = 5000 -- number of solutions to check before reporting back


All of these functions assume that constraints are SORTED. This is used to  
"jump ahead" to the next solution that would meet constraints early-on in the 
depth-first traversal. We give preference to changing indices further right 
in the sequence. It is assumed that the constraints are fully sorted (inner 
and outer) by doing the following for k :: Constraints:

	sort (map sort k)
	
If this is not done, then the constraints cannot be guaranteed to be satisfied.
The findSoln function looks for the first solution satisfying two types of 
constraints: those from let-in statements and more generic, progression-level
constraints supplied as a Predicate function.

> findSoln :: (Eq a, Show a) => 
>     Constraints -> Predicate [a] -> [[a]] -> (Int, [a])
> findSoln k f ecs = 
>     let n = length ecs
>         bs = map (`elem` freeInds n k) [0..length ecs-1]
>         lens = map length ecs 
>         g i = zip3 bs i lens
>         frec j iCurr = 
>             let iNext = findNext2 k $ g iCurr 
>                 soln = zipWith (!!) ecs iCurr
>             in  if last iCurr < 0 then error "No more solutions." else
>                 if f soln then (j, soln) else frec (j+1) iNext
>     in  frec 0 $ take (length ecs) $ repeat 0 

> findSoln2 :: (NFData a, Eq a, Show a) => 
>     Constraints -> Predicate [a] -> [[a]] -> IO (Int, [a])
> findSoln2 k f ecs = 
>     let initVal = take (length ecs) $ repeat 0
>         bs = map (`elem` freeInds (length ecs) k) [0..length ecs-1]
>         lens = map length ecs 
>         g i = zip3 bs i lens
>         frec j iCurr = do
>             let iNext = findNext2 k (g iCurr)
>                 soln = makeSoln 0 ecs iCurr
>                 jNext = j+1
>             if last iCurr < 0 then error "No more solutions." else
>                 if f (force soln) then return (j, soln) else -- SEE NOTE BELOW
>                     if j `mod` printIter == 0 && printIt then 
>                             putStrLn ("Solutions examined: "++show j) >> 
>                             frec jNext iNext
>                     else frec jNext iNext
>     in  frec 0 initVal 

Note on use of force in findSoln2: if force is not used and used exactly where
it has been placed above, the function experiences a thunk leak. Printing iCurr, 
iNext, or soln can help to knock back memory usage, but it still leaks until 
printed unless force is used. Using seq can slow the leak to a trickle, but it 
still occurs and will cause long searches to crash. Using force appears to be
the best solution for this leak.


> findSolnPar :: (NFData a, Eq a, Show a) => 
>     Constraints -> Predicate [a] -> [[a]] -> Int -> IO (Int, [a])
> findSolnPar k f ecs parSize = 
>     let initVal = take (length ecs) $ repeat 0
>         bs = map (`elem` freeInds (length ecs) k) [0..length ecs-1]
>         lens = map length ecs 
>         g i = zip3 bs i lens
>         frec j iCurr = do
>             let iCurrs = filter ((>=0).last) (iCurr : findNextI k (g iCurr) (parSize - 1))
>                 iNext = findNext2 k $ g $ last iCurrs -- for next iteration
>                 solns = filter f $ seq (force iCurrs) 
>                         (map (makeSoln 0 ecs) iCurrs `using` parList rdeepseq)
>                 jNext = j+length iCurrs
>             if last iCurr < 0 then error "No more solutions." else
>                 if not $ null solns then return (j, head solns) else
>                     if printIt then 
>                         putStrLn ("Solutions examined: "++show j) >> frec jNext iNext
>                     else frec jNext iNext
>     in  frec 0 initVal


> makeSoln :: Int -> [[a]] -> [Int] -> [a]
> makeSoln j [] [] = []
> makeSoln j [] is = error "(makeSoln) Not enough equivalence relations!"
> makeSoln j ecs [] = error "(makeSoln) Not enough indices!"
> makeSoln j (e:ecs) (i:is) = 
>     if i < length e then (e !! i) : makeSoln (j+1) ecs is
>     else error ("Bad index at position "++show j++": "++show i++", classes="++show (length e))

> fetch xs i = if i >= length xs then error "(fetch) Index is too large!" else xs !! i


The findNext function performs the actual traversal of the space. It bypasses
a lot of irrelevant points by only looking at points that at least satisfy 
the let-in constraints. Those points may or may not satisfy the Predicate, f, 
and so the function may have to explore many solutions.


> findNext :: Constraints -> [Index] -> [Bound] -> [Index]
> findNext k is lens = 
>     let bs = map (`elem` freeInds (length is) k) [0..length is-1]
>         xs = zip3 bs (is) (lens)
>     in  foldl applyCons (incr xs) k where


This version is used for one of the recursive implementations.

> findNext2 :: Constraints -> [(Bool, Index, Bound)] -> [Int]
> findNext2 k xs = foldl' applyCons (incr2 xs) k

> findNextI :: Constraints -> [(Bool, Index, Bound)] -> Int -> [[Index]]
> findNextI k xs i = 
>     let xs' = findNext2 k xs
>         fixWith = zipWith (\(a,b,c) d -> (a,d,c))
>     in  xs' : take (i-1) (iterate (findNext2 k . fixWith xs) xs') where 

The applyCons function applys let-in constraints to an index list. Indices
on the left are given preference when satisfying constraints.

> applyCons :: [Int] -> [(Int, Int)] -> [Int]
> applyCons inds [] = inds
> applyCons inds ((i,j):ijs) = 
>     foldl (f val) inds (map fst ijs) where -- might need seq here
>     val = (take (j-i+1) $ drop i inds)
>     f val src i = take i src ++ val ++ 
>                   drop (i + length val) src
	
Solutions are explored by keeping track of a list of indices into equivalence
classes. This list is incremented by considering which indices can actually 
be changed (those unconstrained by let-in expressions). If the end of the 
search space is reached, an error is thrown since there is no solution.	



The following modification avoids a memory error if used for traversal
with something like iterate. It returns a "flag" at the end (-1) if 
no solutions exist. This must be picked up by the calling functions for
termination to occur.

> incr :: [(Bool, Int, Int)] -> [Int]
> incr ((b,i,l):xs) = let is = map (\(x,y,z) -> y) xs in
>     if b then if i >= l-1 then 0 : incr xs else (i+1) : is
>     else i : incr xs
> incr [] = error "No more solutions!"


The following version uses a flag instead of an error message:

> incr2 :: [(Bool, Int, Int)] -> [Int]
> incr2 ((b,i,l):xs) = let is = map (\(x,y,z) -> y) xs in
>     if b then if i >= l-1 then 0 : incr2 xs else (i+1) : is
>     else i : incr2 xs
> incr2 [] = [-1] -- no more solutions exist


The freeInds function turns let-in constraints into a list of indices that
can be updated or incremented.

> freeInds :: Int -> Constraints -> [Int]
> freeInds n k = 
>     let k' = map (map (\(i,j) -> [i..j])) k 
>         t = nub $ concat $ concatMap tail k'
>     in  filter (not . (`elem` t)) [0..n-1] 


==================================


