Post Processing Module to Link Grammar with OPTIC Functions
Donya Quick and Paul Hudak
Last modified: 14-June-2013
For paper: Grammar-Based Automated Music Composition in Haskell

Post processing module to turn Terms into music using Euterpea.

> module PostProc where
> import Euterpea hiding (transpose)
> import GrammarMonad
> import ChordSpaces
> import Data.List
> import System.Random

Intermediate types:

> type Key = (AbsPitch, Mode)
> type TChord = (Key, Dur, AbsChord)
> type TNote = (Key, Dur, AbsPitch)
> type Voice = [TNote]


The goal using these intermediate types is the following:

INPUT	     STEP            OUTPUT         FUNCTION
Seeds -----(grammar)-------> Sentence       gen
Sentence --(mode info)-----> [TChord]       toAbsChords
[TChord] ------------------> [Voice] 		toVoices
[Voice] -------------------> Music Pitch    vsToMusic or vsToMusicI


Function to convert into intermediate type [TChord]:

> toAbsChords :: Term -> Mode -> [TChord]
> toAbsChords (NT (Chord d c)) m = [((0, m), d,toAs c m)]
> toAbsChords (S ts) m = concatMap (\t -> toAbsChords t m) ts
> toAbsChords (Mod mt x) m = 
>     let (amt, mt') = getMods mt m
>         f = map (\((k,m),d,c) -> ((k+amt,mt'),d, t amt c)) 
>     in  f (toAbsChords x mt') 
> toAbsChords x m = 
>     error ("(toAbsChords) Unable to handle expression: "++showTerm x)

The toCords functon does a similar thing, but returns a CType and 
its key/mode context without performing the conversion to AbsChord.

> toChords :: Term -> Mode -> [(Key, Dur, CType)]
> toChords (NT (Chord d c)) m = [((0, m), d, c)]
> toChords (S ts) m = concatMap (\t -> toChords t m) ts
> toChords (Mod mt x) m = 
>     let (amt, mt') = getMods mt m
>         f = map (\((k,m),d,c) -> ((k+amt,mt'),d,c)) 
>     in  f (toChords x mt') 
> toChords x m = 
>     error ("(toAbsChords) Unable to handle expression: "++showTerm x)

> procTChords :: [TChord] -> [TChord]
> procTChords tcs = map (\(m,d,ps) -> (m, d, map (`mod` 12) ps)) tcs

> cChordToA :: (Key, Dur, CType) -> TChord
> cChordToA ((k,m),d,ct) = ((k,m), d, t k $ toAs ct m)


Conversion of a single chord to a mode rooted at zero:

> toAs :: CType -> Mode -> [AbsPitch]
> toAs ct m = 
>     let s = getScale m ++ map (+12) s -- fininite scale
>         i = head $ findIndices (==ct) [I, II, III, IV, V, VI, VII] -- can be updated w/enum
>     in  map (s !!) $ map (+i) [0,2,4]

Fundtions to get information on modes:

> getScale :: Mode -> [AbsPitch]
> getScale Major = [0,2,4,5,7,9,11]
> getScale Minor = [0,2,3,5,7,8,10]

> relPat :: Mode -> [Mode]
> relPat Major = [Major, Minor, Minor, Major, Major, Minor, Minor]
> relPat Minor = [Minor, Minor, Major, Minor, Minor, Major, Major]

> getMods mt m = 
>     let mts = [M2, M3, M4, M5, M6, M7]
>         i = (head $ findIndices (==mt) mts)+1
>     in  (getScale m !! i, relPat m !! i)


Conversion of intermediate type to Music Pitch:

> tChordsToMusic :: [TChord] -> Music Pitch
> tChordsToMusic ts = line $ map f ts where
>     f ((k,m),d, as) = chord $ map (\a -> note d (pitch a)) as


============ SPLITTING VOICES APART ===========

The code here places TChords into a form more suitable
for additional musical processing. A Voice is a list of
pitches with duration and key/mode context. 

> toVoices :: [TChord] -> [Voice]
> toVoices ts = 
>     let (ks,ds,ps) = unzip3 ts
>     in  map (\v -> zip3 ks ds v) $ transpose ps

> toNotes :: Voice -> Music Pitch
> toNotes = line . map (\(k,d,p) -> note d (pitch p))

> vsToMusic :: [Voice] -> Music Pitch
> vsToMusic = 
>     chord . map toNotes 

> vsToMusicI :: [InstrumentName] -> [Voice] -> Music Pitch
> vsToMusicI is = 
>     chord . zipWith (\i m -> instrument i m) is . map toNotes


=========

DISPLAY FUNCTIONS

> cShow = latexShow -- preferred format for showing results

> latexShow t ct = show ct ++ "^{" ++ showDur t ++ "}"
> printShow t ct = '(' : show ct ++ " " ++ showDur t ++ ")"

> showTerm :: Term -> String
> showTerm t = case t of 
>     Let x a exp -> "Let "++x++"="++show a++" in "++showTerm exp
>     Mod m t -> show m ++ "( "++ showTerm t ++" )"
>     S s -> (concat $ intersperse " " $ map showTerm s)
>     Var v -> v
>     NT (Chord t ct) -> cShow t ct

> showDur :: Dur -> String
> showDur d = if d == sn then "s" else
>             if d == en then "e" else
>             if d == qn then "q" else
>             if d == hn then "h" else
>             if d == wn then "w" else show d


===================

CONSTRAINTS

Let-in expressions impose constraints on the interpretation of
the results as music. These can be viewed as progression-level
constraints, or predicates.

mkCons fundtion looks through the let-in structure of a Term
and produces a progression-level predicate for use with the
ChordSpaces module.

> mkCons :: (Eq a) => Term -> Predicate [a]
> mkCons t xs = toCons (findInds [] t) xs where
>     toCons :: (Eq a) => [[(Int, Int)]] -> [a] -> Bool
>     toCons [] xs = True
>     toCons (c:cs) xs = 
>         let f (i,j) = take (j+1-i) $ drop i xs 
>         in  (and $ map (f (head c) == ) $ 
>             map f $ tail c) && toCons cs xs


The findInds function looks through a Term for let-in expressions.
When it finds Let x a exp, it calls findIndsSub on x and exp to 
determine which indices in the sequence are occupied by instances
of x. 

> findInds :: [(String, Term)] -> Term -> [[(Int, Int)]]
> findInds e t = case t of
>     Mod m t -> findInds e t -- no extra offsets to worry about
>     S [] -> []
>     S (s:ss) -> 
>            let vs1 = findInds e s
>                len = length $ toAbsChords (expand e s) Major
>                vs2 = findInds e (S ss)
>            in  vs1 ++ map (add' len) vs2 where
>                add' y = map (\(a,b) -> (a+y, b+y))
>     --S s -> let vs = map (findInds e) s
>     --           lens = 0 : map (\v -> length $ toAbsChords (expand e v) Major) s
>     --       in  concat $ zipWith f lens vs where
>     --       f x vs = map (\v -> map (\(a,b) -> (a+x, b+x)) v) vs
>     Let x a exp -> 
>         let e' = (x, expand e a):e
>             xCases = findIndsSub x e' exp
>         in  xCases : findInds e' a ++ findInds e' exp
>     _ -> [] 

The findIndsSub function looks for instances of a variable and
determines what indices they occupy.

> findIndsSub :: String -> [(String, Term)] -> Term -> [(Int, Int)]
> findIndsSub x e t = case t of
>     Let x' a exp -> 
>         let a' = expand e a -- to figure out sequence length
>         in  if x'==x then [] -- x has been redefined, no instances to see here
>             else findIndsSub x e a ++ findIndsSub x ((x',a'):e) exp
>     Var x' -> if x/=x' then [] else 
>         let v = maybe (error (x ++ " is undefined")) id $ lookup x e
>             len = length $ toAbsChords v Major
>         in  [(0, len-1)]
>     Mod m t' -> findIndsSub x e t' -- mods don't change dependencies of offsets
>     S [] -> [] -- no instances to worry about
>     S (s:ss) -> 
>         let len = length $ toAbsChords (expand e s) Major
>         in  findIndsSub x e s ++ add' (len) (findIndsSub x e (S ss)) where
>         add' y = map (\(a,b) -> (a+y, b+y))
>     NT v -> [] -- not a variable, so no instances to worry about




> flatten xs = if flattenB xs == xs then xs else flatten (flattenB xs)

> flattenB t = case t of
>     Let x a exp -> Let x (flattenB a) (flattenB exp)
>     Var x -> Var x
>     Mod m t -> Mod m (flattenB t)
>     S [] -> S []
>     S (S xs1 : xs2) -> flattenB (S (xs1++xs2))
>     S (x : xs) -> S (x : map flattenB xs)
>     NT v -> NT v


Replaces all instances of a variable (allows lexical scoping).

> replaceT :: String -> String -> Term -> Term
> replaceT oldX newX t = case t of 
>     Let x a exp -> if x==oldX then t  -- can't continue, oldX was redefined
>                    else Let x (replaceT oldX newX a) (replaceT oldX newX exp)
>     S s -> S $ map (replaceT oldX newX) s
>     Mod m t' -> Mod m $ replaceT oldX newX t'
>     Var x -> if x==oldX then Var newX else Var x
>     x -> x -- chord case


Converts variable names to unique identifiers.

> toUniques :: Int -> Term -> (Int, Term)
> toUniques i t = case t of 
>     Let x a exp -> -- need to find instances in a and exp
>         let x' = show i
>             (i1,a') = toUniques (i+1) (replaceT x x' a)
>             (i2,exp') = toUniques i1 (replaceT x x' exp)
>         in  (max i1 i2, Let x' a' exp')
>     S s -> let s' = foldl f (i, []) s in (fst s', S $ snd s') where
>         f :: (Int, [Term]) -> Term -> (Int, [Term])
>         f (i, ts) t = let (i', t') = toUniques i t in (max i i', t':ts)
>     Mod m t' -> let (i', t'') = toUniques i t' in (i', Mod m t'')
>     x -> (i, x) -- nothing to change for vars and chords

> toUniques' = snd . toUniques 0


> tag :: [(String, Term)] -> String -> Term -> [String] 
> tag e x t = case t of 
>     Let x' a exp -> tag ((x', expand e a):e) x exp
>     Var x'       -> undefined -- maybe (error (x ++ " is undefined")) (toStrs x) (lookup x e)
>     S s         -> concatMap (tag e x) s
>     Mod m t'    -> tag e x t'
>     NT v           -> undefined -- ["v"]

> toStrs x t = take (length $ toAbsChords t Major) (repeat x)


============

TESTING


> terms = Let "x" (S [Mod M2 (S [Mod M5 (NT (Chord (0.25) I)),NT (Chord (0.25) I)]),Mod M5 (S [NT (Chord (0.25) V),NT (Chord (0.25) I)]),S [S [NT (Chord (0.25) IV),Mod M5 (NT (Chord (0.25) I))],Let "x" (NT (Chord (0.25) I)) (S [Var "x",Var "x"])]]) (S [Var "x",S [Mod M7 (Let "x" (Let "x" (Mod M5 (NT (Chord (0.25) I))) (S [Var "x",S [Mod M7 (NT (Chord (0.25) VII)),Mod M6 (NT (Chord (0.25) I))],Var "x"])) (S [Var "x",Var "x"])),Mod M6 (Let "x" (S [Mod M5 (S [Mod M5 (NT (Chord (0.25) I)),NT (Chord (0.25) I)]),S [Mod M3 (NT (Chord (0.25) I)),NT (Chord (0.25) I)]]) (S [Var "x",Var "x"]))],Var "x"])

terms = (S [Mod M2 (S [Mod M5 (NT (Chord (0.25) I)),NT (Chord (0.25) I)]),Mod M5 (S [NT (Chord (0.25) V),NT (Chord (0.25) I)]),S [S [NT (Chord (0.25) IV),Mod M5 (NT (Chord (0.25) I))],Let "x" (NT (Chord (0.25) I)) (S [Var "x",Var "x"])]]) 

> cons = findInds [] $ terms -- toUniques' $ flatten terms
